import shlex

from mazikeen.GeneratorUtils import getYamlString, getYamlInt, getYamlBool
from mazikeen.GeneratorException import GeneratorException
from mazikeen.DiffBlock import DiffBlock
from mazikeen.Utils import diffStrategy
from mazikeen.GeneratorUtils import getYamlString, getYamlList
from mazikeen.ConsolePrinter import Printer, BufferedPrinter

def _parseStrategy(data, line, field):
    strStrat = getYamlString(data, line, field)
    if not any(x for x in diffStrategy if x.name == strStrat):
        raise GeneratorException(f"Invalid value '{strStrat}' for field '{field}' at line {line}")
    
    return diffStrategy[data]

def _parseignore(data, line, field):
    listignore = getYamlList(data, line, field)
    for ignoreline in listignore:
        if (not isinstance(ignoreline, str)):
            raise GeneratorException(f"Field '{field}' expects a list of strings at line {line}")
    return listignore

def getYamlPaths(data, line = None, field = None):
    if isinstance(data, str):
        dirs = shlex.split(data)
        if (len(dirs) == 2): 
            return dirs
    if (line and field):
        raise GeneratorException(f"Field '{field}' expects two paths at line {line}")
    else:
        raise GeneratorException(f"'diff' block not recognized")

def generateDiffBlock(data):
    if isinstance(data, str):
        return DiffBlock(paths = getYamlPaths(data))
    if not isinstance(data, dict):
        raise DiffBlock(f"'diff' block not recognized")
    args = {}
    key = ""
    knownkeys = {'paths': lambda _data: getYamlPaths(_data, data['__line__'], key),
                 'binarycompare': lambda _data: getYamlBool(_data, data['__line__'], key),
                 'strategy': lambda _data: _parseStrategy(_data, data['__line__'], key),
                 'ignore': lambda _data: _parseignore(_data, data['__line__'], key)}
    for key in data:
        if key == "__line__": continue
        if not key.lower() in knownkeys.keys():
            raise GeneratorException(f"Only one of the following keys are allowed: {[*knownkeys.keys()]} at line {data['__line__']}")
        args[key.lower()] = knownkeys[key.lower()](data[key])
    return DiffBlock(**args)


import unittest
import os
import io
import sys
from mazikeen.ConsolePrinter import Printer, BufferedPrinter

class ConsolePrinterTest(unittest.TestCase):
    def test_basic(self):
        capturedOutput = io.StringIO()
        sys.stdout = capturedOutput
        
        printer = Printer()
        printer.print("Hello World")
        printer.verbose("How do you do?")
        
        self.assertEqual("Hello World\n", capturedOutput.getvalue())
        
        capturedOutput = io.StringIO()
        sys.stdout = capturedOutput
        
        printer = Printer(verbose=True)
        printer.print("Hello World")
        printer.verbose("How do you do?")
        
        self.assertEqual("Hello World\nHow do you do?\n", capturedOutput.getvalue())
        
        sys.stdout = sys.__stdout__

    def test_basicBufferedPrinter(self):
        capturedOutput = io.StringIO()
        sys.stdout = capturedOutput
        
        printer = BufferedPrinter()
        printer.print("Hello World")
        printer.verbose("How do you do?")
        
        self.assertEqual("", capturedOutput.getvalue())
        
        printer.flush()
        self.assertEqual("Hello World\n", capturedOutput.getvalue())
        
        capturedOutput = io.StringIO()
        sys.stdout = capturedOutput
        
        printer = BufferedPrinter(verbose=True)
        printer.print("Hello World")
        printer.verbose("How do you do?")
        
        self.assertEqual("", capturedOutput.getvalue())
        
        printer.flush()
        self.assertEqual("Hello World\nHow do you do?\n", capturedOutput.getvalue())
        
        sys.stdout = sys.__stdout__
        
    def test_getBufferedPrinter(self):
        capturedOutput = io.StringIO()
        sys.stdout = capturedOutput
        
        printer = Printer(verbose = True)
        printer.print("Hello World")
        printer.verbose("How do you do?")
        
        self.assertEqual("Hello World\nHow do you do?\n", capturedOutput.getvalue())
        
        capturedOutput = io.StringIO()
        sys.stdout = capturedOutput
        
        printer = printer.getBufferedPrinter()
        printer.print("Hello World")
        printer.verbose("How do you do?")
        
        self.assertEqual("", capturedOutput.getvalue())
        
        printer.flush()
        self.assertEqual("Hello World\nHow do you do?\n", capturedOutput.getvalue())
        
        capturedOutput = io.StringIO()
        sys.stdout = capturedOutput
        
        printer = printer.getBufferedPrinter()
        printer.print("Hello World")
        printer.verbose("How do you do?")
        
        self.assertEqual("", capturedOutput.getvalue())
        
        printer.flush()
        self.assertEqual("Hello World\nHow do you do?\n", capturedOutput.getvalue())
        
        sys.stdout = sys.__stdout__

    def test_printSeparator(self):
        capturedOutput = io.StringIO()
        sys.stdout = capturedOutput
        
        printer = Printer(verbose = True)
        printer.print("Hello", "World", sep = ",")
        printer.verbose("How", "do", "you", "do?", sep = ",")
        
        self.assertEqual("Hello,World\nHow,do,you,do?\n", capturedOutput.getvalue())
        
        capturedOutput = io.StringIO()
        sys.stdout = capturedOutput
        
        printer = printer.getBufferedPrinter()
        printer.print("Hello", "World", sep = ",")
        printer.verbose("How", "do", "you", "do?", sep = ",")
        printer.flush()
        self.assertEqual("Hello,World\nHow,do,you,do?\n", capturedOutput.getvalue())
        sys.stdout = sys.__stdout__
        
    def test_basicError(self):
        capturedOutput = io.StringIO()
        sys.stdout = capturedOutput
        
        printer = Printer()
        printer.error("Test error")
        
        self.assertEqual("Error: Test error\n", capturedOutput.getvalue())
        
        capturedOutput = io.StringIO()
        sys.stdout = capturedOutput
        
        printer = BufferedPrinter()
        printer.error("Test error")
        
        self.assertEqual("", capturedOutput.getvalue())
        printer.flush()
        
        self.assertEqual("Error: Test error\n", capturedOutput.getvalue())
        sys.stdout = sys.__stdout__
        
if __name__ == '__main__':
    unittest.main()




from mazikeen.GeneratorException import GeneratorException

def getYamlInt(data, line, field):
    if data == None: return None
    if (not isinstance(data, int)):
        if (isinstance(data, dict)):
            raise GeneratorException(f"field '{field}' expects an integer at line {data['__line__']}")
        raise GeneratorException(f"field '{field}' expects an integer at line {line}")

    return data

def getYamlBool(data, line, field):
    if (not isinstance(data, bool)):
        if (isinstance(data, dict)):
            raise GeneratorException(f"field '{field}' expects a bool at line {data['__line__']}")
        raise GeneratorException(f"field '{field}' expects a bool at line {line}")
    return data

def getYamlString(data, line, field):
    if (not isinstance(data, str)):
        if (isinstance(data, dict)):
            raise GeneratorException(f"field '{field}' expects an integer at line {data['__line__']}")
        raise GeneratorException(f"field '{field}' expects a string at line {line}")
    return data

def getYamlList(data, line, field):
    if (not isinstance(data, list)):
        if (isinstance(data, dict)):
            raise GeneratorException(f"field '{field}' expects a list at line {data['__line__']}")
        raise GeneratorException(f"field '{field}' expects a list at line {line}")
    return data


from mazikeen.GeneratorUtils import getYamlString, getYamlInt
from mazikeen.RunBlock import RunBlock
from mazikeen.GeneratorException import GeneratorException

def generateRunBlock(data):
    if isinstance(data, str):
        return RunBlock(cmd = data)
    if not isinstance(data, dict):
        raise GeneratorException(f"'run' block not recognized")
    args = {}
    key = ""
    knownkeys = {'cmd': lambda _data: getYamlString(_data, data['__line__'], key), 
                 'outputfile': lambda _data: getYamlString(_data, data['__line__'], key),
                 'inputfile': lambda _data: getYamlString(_data, data['__line__'], key),
                 'exitcode': lambda _data: getYamlInt(_data, data['__line__'], key),
                 'shell': lambda _data: getYamlString(_data, data['__line__'], key)}
    for key in data:
        if key == "__line__": continue
        if not key.lower() in knownkeys.keys():
            raise GeneratorException(f"Only one of the following keys are allowed: {[*knownkeys.keys()]} at line {data['__line__']}")
        args[key.lower()] = knownkeys[key.lower()](data[key])
    return RunBlock(**args)


import unittest
import time
from concurrent.futures import ThreadPoolExecutor
from mazikeen.ConsolePrinter import Printer, BufferedPrinter
from mazikeen.Loopers import Serial, Parallel

class SerialTest(unittest.TestCase):
    def test_basic(self):
        class MockExecuter:
            def run(self, workingDir, variables = {}, printer = Printer()):
                self.executed = True
                self.workingDir = workingDir
                self.printer = printer
                return True
        
        mockExecuter = MockExecuter()
        loop = Serial([mockExecuter])
        self.assertTrue(loop.run(workingDir = ".", printer = "Dummy Printer"))
        self.assertTrue(hasattr(mockExecuter, 'executed'))
        self.assertTrue(mockExecuter.workingDir == ".")
        self.assertTrue(mockExecuter.printer == "Dummy Printer")

    def test_failfast(self):
        class MockExecuter:
            def __init__(self, res):
                self.res = res

            def run(self, workingDir = ".", variables = {}, printer = Printer()):
                self.executed = True
                return self.res

        mockExecuters = [MockExecuter(True), MockExecuter(False), MockExecuter(True)]
        loop = Serial(mockExecuters, failfast = True)
        self.assertFalse(loop.run(workingDir = ".", printer = Printer()))
        self.assertTrue(hasattr(mockExecuters[0], 'executed'))
        self.assertTrue(hasattr(mockExecuters[1], 'executed'))
        self.assertFalse(hasattr(mockExecuters[2], 'executed'))

        mockExecuters = [MockExecuter(True), MockExecuter(False), MockExecuter(True)]
        loop = Serial(mockExecuters, failfast = False)
        self.assertFalse(loop.run())
        self.assertTrue(hasattr(mockExecuters[0], 'executed'))
        self.assertTrue(hasattr(mockExecuters[1], 'executed'))
        self.assertTrue(hasattr(mockExecuters[2], 'executed'))

    def test_entries(self):
        class MockExecuter:
            def __init__(self):
                self.variables = []

            def run(self, workingDir, variables, printer = Printer()):
                self.variables.append(variables)
                self.workingDir = workingDir
                self.printer = printer
                return True
        
        mockExecuter = MockExecuter()
        loop = Serial([mockExecuter], entries = [{"var1" : "val1"}, {"var2" : "val2"}])
        loop.run(workingDir = ".", printer = "Dummy printer")
        self.assertTrue(mockExecuter.variables == [{"var1" : "val1"}, {"var2" : "val2"}])
        self.assertTrue(mockExecuter.workingDir == ".")
        self.assertTrue(mockExecuter.printer == "Dummy printer")

    def test_VariablesOverwrite(self):
        class MockExecuter:
            def run(self, workingDir, variables, printer = Printer()):
                self.variables = variables
                self.workingDir = workingDir
                return True
        
        mockExecuter = MockExecuter()
        loop = Serial([mockExecuter], entries = [{"var1" : "val1", "var2" : "val2"}])
        loop.run(workingDir = ".", variables = {"var1" : "overwriteVal1", "var3": "val3"})
        self.assertTrue(mockExecuter.variables == {"var1" : "val1", "var2" : "val2",  "var3": "val3"})
        self.assertTrue(mockExecuter.workingDir == ".")

class ParallelTest(unittest.TestCase):
    def test_basic(self):
        class MockExecuter:
            def run(self, workingDir, variables = {}, printer = Printer()):
                self.executed = True
                self.workingDir = workingDir
                self.printer = printer
                return True
        
        mockExecuter = MockExecuter()
        loop = Parallel([mockExecuter], max_workers=4)
        self.assertTrue(loop.run(".", printer = Printer()))
        self.assertTrue(hasattr(mockExecuter, 'executed'))
        self.assertTrue(mockExecuter.workingDir, '.')
        self.assertTrue(isinstance(mockExecuter.printer, BufferedPrinter))

    def test_failfast(self):
        class MockExecuter:
            def __init__(self, res):
                self.res = res

            def run(self, workingDir, variables = {}, printer = Printer()):
                self.executed = True
                return self.res
        
        mockExecuters = [MockExecuter(True), MockExecuter(False), MockExecuter(True)]
        loop = Parallel(mockExecuters, failfast = True, max_workers = 1)
        self.assertFalse(loop.run("."))
        self.assertTrue(hasattr(mockExecuters[0], 'executed'))
        self.assertTrue(hasattr(mockExecuters[1], 'executed'))
        self.assertFalse(hasattr(mockExecuters[2], 'executed'))

        mockExecuters = [MockExecuter(True), MockExecuter(False), MockExecuter(True)]
        loop = Parallel(mockExecuters, failfast = False, max_workers = 1)
        self.assertFalse(loop.run("."))
        self.assertTrue(hasattr(mockExecuters[0], 'executed'))
        self.assertTrue(hasattr(mockExecuters[1], 'executed'))
        self.assertTrue(hasattr(mockExecuters[2], 'executed'))

    def test_entries(self):
        class MockExecuter:
            def __init__(self):
                self.variables = []

            def run(self, workingDir, variables, printer = Printer()):
                self.variables.append(variables)
                self.workingDir = workingDir
                self.printer = printer
                return True
        
        mockExecuter = MockExecuter()
        loop = Parallel([mockExecuter], entries = [{"var1" : "val1"}, {"var2" : "val2"}])
        self.assertTrue(loop.run(".", printer = Printer()))
        self.assertTrue(mockExecuter.variables == [{"var1" : "val1"}, {"var2" : "val2"}])
        self.assertTrue(mockExecuter.workingDir == ".")
        self.assertTrue(isinstance(mockExecuter.printer, BufferedPrinter))

    def test_VariablesOverwrite(self):
        class MockExecuter:
            def run(self, workingDir, variables, printer = Printer()):
                self.variables = variables
                self.workingDir = workingDir
                return True
        
        mockExecuter = MockExecuter()
        loop = Parallel([mockExecuter], entries = [{"var1" : "val1", "var2" : "val2"}])
        loop.run(workingDir = ".", variables = {"var1" : "overwriteVal1", "var3": "val3"})
        self.assertTrue(mockExecuter.variables == {"var1" : "val1", "var2" : "val2",  "var3": "val3"})
        self.assertTrue(mockExecuter.workingDir == ".")

if __name__ == '__main__':
    unittest.main()



import pathlib

from mazikeen.Utils import replaceVariables, ensure_dir, diffStrategy, diff
from mazikeen.ConsolePrinter import Printer, BufferedPrinter


class DiffBlock:
    def __init__(self, paths, binarycompare  = False, strategy = diffStrategy.All, ignore = []):
        assert len(paths) == 2
        self.paths = paths
        self.binarycompare = binarycompare
        self.strategy = strategy
        self.ignore = ignore

    def run(self, workingDir ="", variables = {}, printer = Printer()):
        printer.verbose("Diff:", self.paths)
        _leftpath = replaceVariables(self.paths[0], variables, printer)
        _rightpath = replaceVariables(self.paths[1], variables, printer)
        _ignore = []
        for ignoreLine in self.ignore:
            _ignoreline = replaceVariables(ignoreLine, variables, printer)
            _ignore.append(_ignoreline)
        workingDirPath = pathlib.Path(workingDir)
        return diff(workingDirPath.joinpath(_leftpath), workingDirPath.joinpath(_rightpath), self.binarycompare, self.strategy, _ignore, printer)



from concurrent.futures import ThreadPoolExecutor
import multiprocessing
import threading
import time

import mazikeen.SignalHandler as SignalHandler
from mazikeen.ConsolePrinter import Printer, BufferedPrinter

class Looper:
    def __init__(self, steps = [], entries = [], failfast = True):
        self.entries = entries
        self.steps = steps
        self.failfast = failfast

class Serial(Looper):
    def __init__(self, steps = [], entries = [], failfast = True):
        super().__init__(steps, entries, failfast)
        
    def _runEntry(self, workingDir, variables, printer):
        res = True
        for executor in self.steps:
            executorRes = executor.run(workingDir = workingDir, variables = variables, printer = printer)
            if (SignalHandler.failFast()): 
                printer.error(SignalHandler.signalName() + " received.")
                return False
            if (executorRes == False):
                if (self.failfast): return False
                else: res = False
        return res
    
    def run(self, workingDir = "", variables = {}, printer = Printer()):
        res = True
        if self.entries:
            for loopEntry in self.entries:
                if (SignalHandler.failFast()): 
                    printer.error(SignalHandler.signalName() + " received.")
                    return False
                callVariables = variables.copy()
                callVariables.update(loopEntry)
                if (not self._runEntry(workingDir, callVariables, printer)): 
                    if (self.failfast): return False
                    else: res = False
        else:
            return self._runEntry(workingDir, variables, printer)
        return res

class Parallel(Looper):
    def __init__(self, steps = [], entries = [], failfast = True, max_workers = multiprocessing.cpu_count()):
        super().__init__(steps, entries, failfast)
        self.max_workers = max_workers
        self.listFutures = []
        self.addFutureLock = threading.Lock()
        self.foundFailure = False
    
    def _runStep(self, step, workingDir, variables, printer):
        overridePrinter = (not isinstance(printer, BufferedPrinter)) and (self.max_workers > 1)
        if overridePrinter:
            printer = printer.getBufferedPrinter()
        res = step.run(workingDir = workingDir, variables = variables, printer = printer)
        if (SignalHandler.failFast()):
            printer.error(SignalHandler.signalName() + " received.")
            for future in self.listFutures: future.cancel()
            res = False
        if overridePrinter: 
            printer.flush() #step printer was created so that messages are not mixed up
        if res == False and self.foundFailure == False:
            with self.addFutureLock:
                self.foundFailure = True
                if (self.failfast):
                    for future in self.listFutures: future.cancel()
        return res

    def _runEntry(self, poolExecutor, workingDir, variables, printer):
        for step in self.steps:
            with self.addFutureLock:
                if ((self.foundFailure and self.failfast) or SignalHandler.failFast()): return
                self.listFutures.append(poolExecutor.submit(self._runStep, step, workingDir = workingDir, variables = variables, printer = printer))
    
    def run(self, workingDir = "", variables = {}, printer = Printer()):
        self.listFutures = []
        with ThreadPoolExecutor(max_workers=self.max_workers) as poolExecutor:
            if self.entries:
                for loopEntry in self.entries:
                    callVariables = variables.copy()
                    callVariables.update(loopEntry)
                    self._runEntry(poolExecutor, workingDir = workingDir, variables = callVariables, printer = printer)
            else:
                self._runEntry(poolExecutor, workingDir = workingDir, variables = variables, printer = printer)
            if threading.current_thread().name == 'MainThread':
                while(poolExecutor._work_queue.qsize()):
                    time.sleep(0.05)
        for future in self.listFutures:
            if future.cancelled() or future.result() == False: return False
        return True


import os
import io
import glob
import yaml
import time
import pathlib
import argparse
import multiprocessing
import copy
import re

from pathlib import Path, PurePath
from junit_xml import TestSuite, TestCase

import mazikeen.SignalHandler as SignalHandler
from mazikeen.GeneratorLooper import generateSerialBlock, generateParallelBlock
from mazikeen.Loopers import Serial, Parallel
from mazikeen.ConsolePrinter import Printer, BufferedPrinter
from mazikeen.version import __version__
from mazikeen.Utils import ensure_dir
from mazikeen.ScriptDataProcessor import processScriptData

def getScriptFiles(dir, scriptName, maxLevel=-1):
    curLevel = 0
    curLevelDirs = [Path(dir)]
    scriptFiles = []
    while(curLevelDirs and (curLevel != maxLevel)):
        nextLevelDir = []
        for curDir in curLevelDirs:
            tmpNextLevelDir = []
            try:
                dirContent = os.listdir(curDir)
            except OSError:
                continue # Skip directories that can't be read
            for curFile in map(lambda path: Path(curDir).joinpath(path),dirContent): 
                if curFile.is_dir(): 
                    tmpNextLevelDir.append(curFile)
                elif Path(curFile).name == scriptName:
                    tmpNextLevelDir.clear()
                    scriptFiles.append(str(curFile))
                    break
            nextLevelDir.extend(tmpNextLevelDir)
        curLevelDirs = nextLevelDir
        curLevel += 1
    scriptFiles.sort()
    return scriptFiles

def createTestSuits(scriptFiles, root):
    lenRootParts = len(Path(root).parts)
    testSuits = []
    for scriptFile in map(Path, scriptFiles):
        tc = TestCase(scriptFile.parent.name, file = scriptFile)
        tsName = scriptFile.parts[-3] if len(scriptFile.parts) - lenRootParts > 2 else ""
        ts = next((testSuit for testSuit in testSuits if testSuit.name == tsName), None)
        if ts == None:
            testSuits.append(TestSuite(tsName, [tc]))
        else:
            ts.test_cases.append(tc)

    return testSuits

def parseArguments():
    parser = argparse.ArgumentParser(description='Mazikeen test enviroment', prog='mazikeen')
     
    parser.add_argument( '-p','--pattern', metavar='PATTERN', type=str,
                        help='Only run tests which match pattern. Does support also negative patterns "-PATTERN"')
    parser.add_argument( '-f','--failfast', action='store_true', 
                        help='stop on first faild test as quickly as possible')
    parser.add_argument( '--upgradeScriptFile', action='store_true', 
                        help='save upgraded script file. Script files are upgraded if their version is lower that latest version')
    parser.add_argument( '--scriptName', metavar='NAME', type=str, default="script.yaml",
                        help='Mazikeen script name (`script.yaml` default)')
    parser.add_argument( '-s','--start-directory', dest='start', metavar='DIR', type=pathlib.Path, default=".",
                         help="Directory to start discovering tests ('.' default)")
    parser.add_argument('-v', '--verbose', dest='verbose', action='count',
                         help='Verbose output')
    parser.add_argument('-j', '--jobs', dest='jobs', type=int, default = None, const = multiprocessing.cpu_count(), nargs='?',
                         help='Specifies the number of jobs to run simultaneously')
    parser.add_argument('-r', '--report', dest='reportfile', type=pathlib.Path,
                         help='Create junit test report')
    parser.add_argument('--version', action='version', version='%(prog)s ' + __version__)

    return parser.parse_args()

def markTestSkipedByPattern(testSuits, pattern):
    if not isinstance(pattern,str): return
    _pattern = copy.copy(pattern)
    isNegativePatern = _pattern.startswith("-")
    if isNegativePatern: _pattern = _pattern[1:]
    try:
        p = re.compile(_pattern)
    except Exception as e:
        print("Invalid --pattern argument \""+pattern+"\":",str(e))
        exit(1)
    for ts in testSuits: 
        for tc in ts.test_cases:
            if ( (p.match(tc.name) != None) == isNegativePatern):
                tc.add_skipped_info("skipped because of --pattern argumet")

def getReport(testSuits, processTime, executionTime):
    tcPassed = 0
    tcSkipped = 0
    tcFailed = 0 
    tcError = 0
    reportTxt = io.StringIO()

    for ts in testSuits:
        for tc in ts.test_cases:
            if tc.is_error(): tcError += 1
            elif tc.is_failure(): tcFailed += 1
            elif tc.is_skipped(): tcSkipped += 1
            else: tcPassed += 1

    print("----------------------------------------------------------------", file = reportTxt)
    print("Total test cases:", tcPassed + tcSkipped + tcError + tcFailed, "passed:", tcPassed, "skipped:", tcSkipped, "error:", tcError, "failed:", tcFailed, file = reportTxt)
    if (tcSkipped > 0):
        print("----------------------------------------------------------------", file = reportTxt)
        print("Skipped:", file = reportTxt)
        for ts in testSuits:
            for tc in ts.test_cases:
                if tc.is_skipped(): print("\t"+ts.name+("/" if ts.name != "" else "")+tc.name, file = reportTxt)

    if (tcError > 0):
        print("----------------------------------------------------------------", file = reportTxt)
        print("Error:", file = reportTxt)
        for ts in testSuits:
            for tc in ts.test_cases:
                if tc.is_error(): print("\t"+ts.name+("/" if ts.name != "" else "")+tc.name, file = reportTxt)

    if (tcFailed > 0):
        print("----------------------------------------------------------------", file = reportTxt)
        print("Failed:", file = reportTxt)
        for ts in testSuits:
            for tc in ts.test_cases:
                if tc.is_failure(): print("\t"+ts.name+("/" if ts.name != "" else "") +tc.name, file = reportTxt)

    print("----------------------------------------------------------------", file = reportTxt)
    print("process time:",  '%.2f' % processTime, "execution time:",  '%.2f' % executionTime, file = reportTxt)

    return {"text" : reportTxt.getvalue(), "passed" : tcPassed, "skipped" : tcSkipped, "error" : tcError, "failed" : tcFailed }

class TestExecuter:
    def __init__(self, testSuite, testCase, upgradeScriptFile):
        self.testSuite = testSuite
        self.testCase = testCase
        self.upgradeScriptFile = upgradeScriptFile

    def run(self, workingDir, variables = {}, printer = Printer()):
        curDir = os.getcwd()
        printer.print("["+"RUN".ljust(10) + "] ---", self.testSuite.name + ("/" if self.testSuite.name != "" else "") + self.testCase.name)
        try:
            startTime = time.time()
            processStartTime = time.process_time()
            data = processScriptData(self.testCase.file, saveUpgradedScript = self.upgradeScriptFile, printer = printer)
            block = generateSerialBlock(data)
            if block == None: return
            res = block.run(workingDir = str(PurePath(self.testCase.file).parent), printer = printer)
            if (not res):
                self.testCase.add_failure_info(printer.errorOutput or "failed")
        except Exception as e:
            self.testCase.add_error_info(str(e))
            printer.print(str(e))
        finally:
            self.testCase.elapsed_sec = time.time() - startTime
            testResStr = "PASSED"
            if self.testCase.is_error(): 
                testResStr = "ERROR"
            elif self.testCase.is_failure(): 
                testResStr = "FAILED"
            printStr = "["+testResStr.rjust(10) + "] --- " + self.testSuite.name + ("/" if self.testSuite.name != "" else "") + self.testCase.name
            if (printer.isVerbose()): 
                printStr += " --- process time: "+  '%.2f' % (time.process_time() - processStartTime) + " execution time: "+  '%.2f' % (time.time() - startTime)
            printer.print(printStr)
        return testResStr == "PASSED"

def main():
    SignalHandler.init()
    args = parseArguments()
    scriptFiles = getScriptFiles(args.start, args.scriptName)
    testSuits = createTestSuits(scriptFiles, args.start)
    markTestSkipedByPattern(testSuits, args.pattern)
    genTestExecuter = [TestExecuter(ts, tc, args.upgradeScriptFile) for ts in testSuits for tc in ts.test_cases if not tc.is_skipped()]
    if (args.jobs == None): 
        runner = Serial(genTestExecuter, failfast=args.failfast)
    else: 
        runner = Parallel(genTestExecuter, failfast=args.failfast, max_workers = args.jobs)
    startTime = time.time()
    processStartTime = time.process_time()
    res = runner.run(".", printer = Printer(args.verbose != None))
    processTime = time.process_time() - processStartTime
    executionTime = time.time() - startTime
    if (args.failfast and res == False) or SignalHandler.failFast(): 
        for ts in testSuits: 
            for tc in ts.test_cases:
                if tc.elapsed_sec == None:
                    if (SignalHandler.failFast()):
                        tc.add_skipped_info("skipped due to " + SignalHandler.signalName())
                    else:
                        tc.add_skipped_info("skipped due to --failfast argument")
    report = getReport(testSuits, processTime, executionTime)
    Printer(args.verbose != None).print(report["text"])
    if (args.reportfile):
        ensure_dir(args.reportfile)
        with open(args.reportfile, "w") as fw:
            fw.write(TestSuite.to_xml_string(testSuits))
    exit(report["error"] + report["failed"] > 0)
    
if __name__ == '__main__':
    main()


import unittest
import io
import os
import sys
import pathlib
from mazikeen.Utils import replaceVariables, diff, diffStrategy
from mazikeen.GeneratorException import GeneratorException
from mazikeen.ConsolePrinter import Printer, BufferedPrinter

class replaceVariables_Test(unittest.TestCase):
    def test_basic(self):
        line = replaceVariables("Replace ${repl1} ${repl2} ${repl3} \${esc}", {'repl1' : '${repl2}', 'repl2' : '${repl3}', 'repl3' : '${repl1}'})
        self.assertEqual(line, "Replace ${repl2} ${repl3} ${repl1} ${esc}")
        
    def test_missingVar(self):
        with self.assertRaisesRegex(GeneratorException, "Variable .*") as e:
            line = replaceVariables("Replace ${repl1} ${repl2} ${repl3} ${repl4}", {'repl1' : '${repl2}', 'repl2' : '${repl3}', 'repl3' : '${repl1}'})

class diff_Test(unittest.TestCase):
    def test_basic(self):
        self.assertTrue(diff("TestFiles/diff_Test/test_basic/cmp1", "TestFiles/diff_Test/test_basic/cmp2"))

    def test_basicFileOnly(self):
        capturedOutput = io.StringIO()
        sys.stdout = capturedOutput
        self.assertTrue(diff("TestFiles/diff_Test/test_basicFileOny/valid1.txt", "TestFiles/diff_Test/test_basicFileOny/valid1.txt"))
        self.assertFalse(diff("TestFiles/diff_Test/test_basicFileOny/invalid1.txt", "TestFiles/diff_Test/test_basicFileOny/invalid2.txt"))
        sys.stdout = sys.__stdout__
        str = capturedOutput.getvalue()
        self.assertEqual(f"Error: diff failed: \'{pathlib.Path('TestFiles/diff_Test/test_basicFileOny/invalid1.txt')}\' != \'{pathlib.Path('TestFiles/diff_Test/test_basicFileOny/invalid2.txt')}\'\nwhere: 1: b'valid' != 1: b'invalid'\n", capturedOutput.getvalue())

    def test_LinuxNWindowsEOL(self):
        self.assertTrue(diff("TestFiles/diff_Test/test_LinuxNWindowsEOL/cmp1", "TestFiles/diff_Test/test_LinuxNWindowsEOL/cmp2"))

    def test_LinuxNWindowsEOLBinary(self):
        capturedOutput = io.StringIO()
        sys.stdout = capturedOutput
        self.assertFalse(diff("TestFiles/diff_Test/test_LinuxNWindowsEOL/cmp1", "TestFiles/diff_Test/test_LinuxNWindowsEOL/cmp2", binaryCompare = True) == None)
        sys.stdout = sys.__stdout__
        expectedPath0 = pathlib.Path(r"TestFiles/diff_Test/test_LinuxNWindowsEOL/cmp1/file1.txt")
        expectedPath1 = pathlib.Path(r"TestFiles/diff_Test/test_LinuxNWindowsEOL/cmp2/file1.txt")
        printStr = capturedOutput.getvalue()
        self.assertEqual(f"Error: diff failed: '{expectedPath0}' != '{expectedPath1}'\nwhere: 1: b'line1\\r\\n' != 1: b'line1\\n'\n", printStr)

    def test_IgnoreLeftOrphans(self):
        capturedOutput = io.StringIO()
        sys.stdout = capturedOutput
        
        self.assertFalse(diff("TestFiles/diff_Test/test_LeftOrphans/cmp1", "TestFiles/diff_Test/test_LeftOrphans/cmp2"))
        self.assertTrue(diff("TestFiles/diff_Test/test_LeftOrphans/cmp1", "TestFiles/diff_Test/test_LeftOrphans/cmp2", diffStrategy = diffStrategy.IgnoreLeftOrphans))
        
        sys.stdout = sys.__stdout__
        
        expectedPath0 = pathlib.Path(r"TestFiles/diff_Test/test_LeftOrphans/cmp1/f2/file3.txt")
        expectedPath1 = pathlib.Path(r"TestFiles/diff_Test/test_LeftOrphans/cmp2/f2")
        self.assertEqual(f"Error: diff failed: '{expectedPath0}' not in '{expectedPath1}'\n", capturedOutput.getvalue())

    def test_IgnoreRightOrphans(self):
        capturedOutput = io.StringIO()
        sys.stdout = capturedOutput
        
        self.assertFalse(diff("TestFiles/diff_Test/test_RightOrphans/cmp1", "TestFiles/diff_Test/test_RightOrphans/cmp2"))
        self.assertTrue(diff("TestFiles/diff_Test/test_RightOrphans/cmp1", "TestFiles/diff_Test/test_RightOrphans/cmp2", diffStrategy = diffStrategy.IgnoreRightOrphans))
        
        sys.stdout = sys.__stdout__
        
        expectedPath0 = pathlib.Path(r"TestFiles/diff_Test/test_RightOrphans/cmp2/f1/file3.txt")
        expectedPath1 = pathlib.Path(r"TestFiles/diff_Test/test_RightOrphans/cmp1/f1")
        self.assertEqual(f"Error: diff failed: '{expectedPath0}' not in '{expectedPath1}'\n", capturedOutput.getvalue())

    def test_IgnoreOrphans(self):
        capturedOutput = io.StringIO()
        sys.stdout = capturedOutput

        self.assertTrue(diff("TestFiles/diff_Test/test_LeftNRightOrphans/cmp1", "TestFiles/diff_Test/test_LeftNRightOrphans/cmp2", diffStrategy = diffStrategy.IgnoreOrphans))

        sys.stdout = sys.__stdout__
        self.assertEqual("", capturedOutput.getvalue())

    def test_fileAsDirectory(self):
        
        capturedOutput = io.StringIO()
        sys.stdout = capturedOutput
        expectedPath0 = pathlib.Path(r"TestFiles/diff_Test/test_fileAsDirectory/cmp1/f2/file3.txt")
        expectedPath1 = pathlib.Path(r"TestFiles/diff_Test/test_fileAsDirectory/cmp2/f2/file3.txt")
        if not os.path.exists(expectedPath0):
            os.makedirs(expectedPath0)
        self.assertFalse(diff("TestFiles/diff_Test/test_fileAsDirectory/cmp1", "TestFiles/diff_Test/test_fileAsDirectory/cmp2"))

        sys.stdout = sys.__stdout__
        self.assertEqual(f"Error: diff failed: '{expectedPath0}' != '{expectedPath1}'\n", capturedOutput.getvalue())

    def test_InvalidPath(self):
        capturedOutput = io.StringIO()
        sys.stdout = capturedOutput

        self.assertFalse(diff("invalidPath1", "invalidPath2"))

        sys.stdout = sys.__stdout__
        expectedPath0 = pathlib.Path(r"TestFiles/diff_Test/test_fileAsDirectory/cmp1/f2/file3.txt")
        expectedPath1 = pathlib.Path(r"TestFiles/diff_Test/test_fileAsDirectory/cmp2/f2/file3.txt")
        self.assertEqual(f"Error: diff failed: 'invalidPath1' doesn't exist\n", capturedOutput.getvalue())

    def test_Ignore(self):
        self.assertTrue(diff("TestFiles/diff_Test/test_Ignore/cmp1", "TestFiles/diff_Test/test_Ignore/cmp2", ignore = ['m','time=.*?\.', 'date and time', '1..31.21']))
    
    def test_Ignore_False(self):
        printer = Printer(verbose = True)
        capturedOutput = io.StringIO()
        sys.stdout = capturedOutput
        self.assertFalse(diff("TestFiles/diff_Test/test_Ignore_False/cmp1", "TestFiles/diff_Test/test_Ignore_False/cmp2", ignore = ['m','time=.*?\.', 'date and time', '1..31.21', "ignoreText"], printer = printer))
        printedStr = capturedOutput.getvalue()
        expectedPath0 = pathlib.Path(r"TestFiles/diff_Test/test_Ignore_False/cmp1/file1.txt")
        expectedPath1 = pathlib.Path(r"TestFiles/diff_Test/test_Ignore_False/cmp2/file1.txt")
        resStr = f"Error: diff failed: '{expectedPath0}' != '{expectedPath1}'\nwhere: 2: b'y  Additional text so that test fails\\r\\n' != 3: b'y \\n'\n"
        self.assertEqual(printedStr, resStr)

if __name__ == '__main__':
    unittest.main()



import unittest
import yaml
from mazikeen.GeneratorLooper import generateSerialBlock
from mazikeen.ScriptDataProcessor import SafeLineLoader
from mazikeen.MakedirsBlock import MakedirsBlock
from mazikeen.GeneratorException import GeneratorException

class GeneratorMakedirsBlock_test(unittest.TestCase):
    def test_basic(self):
        with open('TestFiles/MakedirsBlock_test/test_basic/script.yaml') as f:
            data = yaml.load(f, Loader=SafeLineLoader)
            block = generateSerialBlock(data)
            self.assertTrue(isinstance(block.steps[0], MakedirsBlock))
            self.assertEqual(block.steps[0].dir, 'Output/MkdirBlock_test/TestDir1')

if __name__ == '__main__':
    unittest.main()



import unittest

from mazikeen.ScriptDataProcessor import Version

class VersionTest(unittest.TestCase):
    def test_simple(self):
        v = Version("1.0.0")
        self.assertTupleEqual((1,0,0), (v.major, v.minor, v.patch))

    def test_smallVersion(self):
        v = Version("1")
        self.assertEqual((1,"",""), (v.major, v.minor, v.patch))

    def test_LongVersion(self):
        with self.assertRaises(Exception, msg = "Invalid version. Version is too long") as e:
            Version("1.0.0.0")
        
    def test_StrInVersion(self):
        with self.assertRaises(ValueError, msg = "Invalid version. Version may only contain numbers") as e:
            Version("1.a.0")

    def test_NoneNEmpty(self):
        v = Version("")
        self.assertTupleEqual(("","",""), (v.major, v.minor, v.patch))
        v = Version(None)
        self.assertTupleEqual(("","",""), (v.major, v.minor, v.patch))

    def test_Compare(self):
        self.assertGreater(Version("1.0.1"), Version("1.0.0"))
        self.assertGreater(Version("1.1.0"), Version("1.0.0"))
        self.assertGreater(Version("2.0.0"), Version("1.0.0"))
        self.assertTrue(Version("1.0.0") == Version("1.0.0"))
        self.assertLess(Version("1.0.0"), Version("1.0.1"))
        self.assertLess(Version("1.0.0"), Version("1.1.0"))
        self.assertLess(Version("1.0.0"), Version("2.0.0"))

if __name__ == '__main__':
    unittest.main()


import itertools

from mazikeen.Loopers import Serial, Parallel
from mazikeen.GeneratorException import GeneratorException
from mazikeen.GeneratorRunBlock import generateRunBlock
from mazikeen.GeneratorMakedirs import generateMakedirs
from mazikeen.GeneratorRmdir import generateRmdir
from mazikeen.GeneratorDiffBlock import generateDiffBlock
from mazikeen.GeneratorUtils import getYamlInt, getYamlBool
from mazikeen.ConsolePrinter import Printer, BufferedPrinter

def _getVariables(data):
    dictVar = {}
    for key in data:
        if key == "__line__": continue
        dictVar[key] = data[key]
    return dictVar
def _dict_product(dicts):
    return (dict(zip(dicts, x)) for x in itertools.product(*dicts.values()))
def __dict_zip_longest(dicts):
    return (dict(zip(dicts, x)) for x in itertools.zip_longest(*dicts.values()))
def _dict_zip(dicts):
    return (dict(zip(dicts, x)) for x in zip(*dicts.values()))
def _parseProduct(data):
    dictVar = _getVariables(data)
    return _dict_product(dictVar)
def _parseZip(data):
    dictVar = _getVariables(data)
    for curDic in __dict_zip_longest(dictVar):
        yield { k:v for k, v in curDic.items() if v }
def _parseEntries(data):
    if data == None: return []
    knownkeys = {'product':_parseProduct, 'zip':_parseZip}
    entries = []
    for key in data:
        if key == "__line__": continue
        if not key.lower() in knownkeys.keys():
            raise GeneratorException(f"Only one of the following keys are allowed: {[*knownkeys.keys()]} at line {data['__line__']}")
        entries.extend(knownkeys[key](data[key]))
    return entries
def _parseSteps(data, line):
    steps = list()
    knownkeys = {'serial':generateSerialBlock, 
                 'parallel':generateParallelBlock, 
                 'run':generateRunBlock,
                 'diff':generateDiffBlock,
                 'makedirs':generateMakedirs,
                 'rmdir':generateRmdir}
    for yamlSteps in data:
        for key in yamlSteps:
            if key == "__line__": continue
            if not key.lower() in knownkeys.keys():
                if (isinstance(data, dict)):
                    raise GeneratorException(f"Error: Key '{key}' not allowed. Only one of the following keys are allowed :{[*knownkeys.keys()]} at line {data['__line__']}")
                else:
                    raise GeneratorException(f"Error: Key '{key}' not allowed. Only one of the following keys are allowed :{[*knownkeys.keys()]} at line {line}")
            steps.append(knownkeys[key.lower()](yamlSteps[key]))
    return steps
def generateSerialBlock(data):
    if data == None: return
    args = {}
    knownkeys = {'entries': lambda data: list(_parseEntries(data)),
                 'failfast': lambda _data: getYamlBool(_data, data['__line__'], key),
                 'steps': lambda _data: list(_parseSteps(_data, data['__line__']))}
    for key in data:
        if key == "__line__": continue
        if not key.lower() in knownkeys.keys():
            raise GeneratorException(f"Only one of the following keys are allowed: {[*knownkeys.keys()]} at line {data['__line__']}")
        args[key.lower()] = knownkeys[key.lower()](data[key])
    return Serial(**args)

def generateParallelBlock(data):
    args = {}
    key = ""
    knownkeys = {'entries': lambda data: list(_parseEntries(data)), 
                 'steps': lambda _data: list(_parseSteps(_data, data['__line__'])),
                 'failfast': lambda _data: getYamlBool(_data, data['__line__'], key),
                 'max_workers': lambda _data: getYamlInt(_data, data['__line__'], key)}
    for key in data:
        if key == "__line__": continue
        if not key.lower() in knownkeys.keys():
            raise GeneratorException(f"Only one of the following keys are allowed: {[*knownkeys.keys()]} at line {data['__line__']}")
        args[key.lower()] = knownkeys[key.lower()](data[key])
    return Parallel(**args)


import unittest
import yaml
from mazikeen.GeneratorLooper import generateSerialBlock
from mazikeen.ScriptDataProcessor import SafeLineLoader
from mazikeen.RunBlock import RunBlock
from mazikeen.GeneratorException import GeneratorException

class GeneratorRunBlock_test(unittest.TestCase):
    def test_basic(self):
        with open('TestFiles/GeneratorRunBlock_test/test_basic/script.yaml') as f:
            data = yaml.load(f, Loader=SafeLineLoader)
            block = generateSerialBlock(data)
            self.assertTrue(isinstance(block.steps[0], RunBlock))
            self.assertEqual(block.steps[0].cmd, "echo $shapes $color")
            self.assertEqual(block.steps[0].exitcode, 0)
            self.assertEqual(block.steps[0].inputfile, 'input/inp.txt')
            self.assertEqual(block.steps[0].outputfile, 'output/out.txt')

    def test_simple(self):
        with open('TestFiles/GeneratorRunBlock_test/test_simple/script.yaml') as f:
            data = yaml.load(f, Loader=SafeLineLoader)
            block = generateSerialBlock(data)
            self.assertTrue(isinstance(block.steps[0], RunBlock))
            self.assertEqual(block.steps[0].cmd, "echo $shapes $color")
            self.assertEqual(block.steps[0].exitcode, None)
            self.assertEqual(block.steps[0].inputfile, None)
            self.assertEqual(block.steps[0].outputfile, None)

    def test_exitCode_None(self):
        with open('TestFiles/GeneratorRunBlock_test/test_exitCode_None/script.yaml') as f:
            data = yaml.load(f, Loader=SafeLineLoader)
            block = generateSerialBlock(data)
            self.assertTrue(isinstance(block.steps[0], RunBlock))
            self.assertEqual(block.steps[0].cmd, "echo $shapes $color")
            self.assertEqual(block.steps[0].exitcode, None)
            self.assertEqual(block.steps[0].inputfile, None)
            self.assertEqual(block.steps[0].outputfile, None)
            
    def test_invalid_commad(self):
        with open('TestFiles/GeneratorRunBlock_test/test_invalid_commad/script.yaml') as f:
            data = yaml.load(f, Loader=SafeLineLoader)
            with self.assertRaisesRegex(GeneratorException, "'run' block not recognized") as e:
                block = generateSerialBlock(data)

    def test_invalid_exitcode(self):
        with open('TestFiles/GeneratorRunBlock_test/test_invalid_exitcode/script.yaml') as f:
            data = yaml.load(f, Loader=SafeLineLoader)
            with self.assertRaisesRegex(GeneratorException, "field 'exitCode' expects an integer at line .*") as e:
                block = generateSerialBlock(data)
if __name__ == '__main__':
    unittest.main()




import unittest
import os
import pathlib
from mazikeen.RmdirBlock import RmdirBlock
from mazikeen.ConsolePrinter import Printer
from mazikeen.Utils import ensure_dir

class RmdirsBlockTest(unittest.TestCase):
    def test_basic(self):
        printer = Printer()
        testPath = pathlib.Path('TestFiles/RmdirBlockTest/Output/TestDir1')
        if not os.path.exists(testPath):
            os.makedirs(testPath)
        self.assertTrue(testPath.is_dir())
        block = RmdirBlock('RmdirBlockTest/Output')
        res = block.run(workingDir = "TestFiles")
        self.assertEqual(res, True)
        self.assertFalse(pathlib.Path('TestFiles/Output').is_dir())
    
    def test_basic_withTailoringSlash(self):
        printer = Printer()
        testPath = pathlib.Path('TestFiles/RmdirBlockTest/Output/TestDir1/')
        if not os.path.exists(testPath):
            os.makedirs(testPath)
        self.assertTrue(testPath.is_dir())
        block = RmdirBlock('RmdirBlockTest/Output')
        res = block.run(workingDir = "TestFiles")
        self.assertEqual(res, True)
        self.assertFalse(pathlib.Path('TestFiles/RmdirBlockTest/Output').is_dir())

if __name__ == '__main__':
    unittest.main()




import unittest
import yaml
from mazikeen.ScriptDataProcessor import SafeLineLoader
from mazikeen.GeneratorException import GeneratorException
from mazikeen.GeneratorLooper import generateSerialBlock
from mazikeen.DiffBlock import DiffBlock
from mazikeen.Utils import diffStrategy

class GeneratorDiffBlock_test(unittest.TestCase):
    def test_basic(self):
        with open('TestFiles/GeneratorDiffBlock_test/test_basic/script.yaml') as f:
            data = yaml.load(f, Loader=SafeLineLoader)
            block = generateSerialBlock(data)
            self.assertTrue(isinstance(block.steps[0], DiffBlock))
            self.assertEqual(block.steps[0].binarycompare, True)
            self.assertEqual(block.steps[0].ignore, ['Time:'])
            self.assertEqual(block.steps[0].paths, ['output/leftpath', 'output/rightpath'])
            self.assertEqual(block.steps[0].strategy, diffStrategy.IgnoreLeftOrphans)

    def test_simple_fail(self):
        with open('TestFiles/GeneratorDiffBlock_test/test_simple_fail/script.yaml') as f:
            data = yaml.load(f, Loader=SafeLineLoader)
            with self.assertRaisesRegex(GeneratorException, "'diff' block not recognized") as e:
                block = generateSerialBlock(data)

    def test_invalid_binarycompare(self):
        with open('TestFiles/GeneratorDiffBlock_test/test_invalid_binarycompare/script.yaml') as f:
            data = yaml.load(f, Loader=SafeLineLoader)
            with self.assertRaisesRegex(GeneratorException, "field 'binarycompare' expects a bool at line .*") as e:
                block = generateSerialBlock(data)

    def test_invalid_strategy(self):
        with open('TestFiles/GeneratorDiffBlock_test/test_invalid_strategy/script.yaml') as f:
            data = yaml.load(f, Loader=SafeLineLoader)
            with self.assertRaisesRegex(GeneratorException, "Invalid value 'IgnoreLeftOrphanss' for field 'strategy' at line .*") as e:
                block = generateSerialBlock(data)

    def test_invalid_ignore(self):
        with open('TestFiles/GeneratorDiffBlock_test/test_invalid_ignore/script.yaml') as f:
            data = yaml.load(f, Loader=SafeLineLoader)
            with self.assertRaisesRegex(GeneratorException, "field 'ignore' expects a list at line .*") as e:
                block = generateSerialBlock(data)
                
    def test_invalid_paths_only_one_path(self):
        with open('TestFiles/GeneratorDiffBlock_test/test_invalid_paths_only_one_path/script.yaml') as f:
            with self.assertRaisesRegex(GeneratorException, "Field 'paths' expects two paths at line 4") as e:
                data = yaml.load(f, Loader=SafeLineLoader)
                block = generateSerialBlock(data)
            
    def test_invalid_paths_three_paths(self):
        with open('TestFiles/GeneratorDiffBlock_test/test_invalid_paths_three_paths/script.yaml') as f:
            with self.assertRaisesRegex(GeneratorException, "Field 'paths' expects two paths at line 4") as e:
                data = yaml.load(f, Loader=SafeLineLoader)
                block = generateSerialBlock(data)
if __name__ == '__main__':
    unittest.main()



import copy
import yaml
from yaml.loader import SafeLoader

from mazikeen.ConsolePrinter import Printer

class SafeLineLoader(SafeLoader):
    def construct_mapping(self, node, deep=False):
        mapping = super(SafeLineLoader, self).construct_mapping(node, deep=deep)
        # Add 1 so line numbering starts at 1
        mapping['__line__'] = node.start_mark.line + 1
        return mapping

class Version():
    def __init__(self, versionStr): 
        (self.major, self.minor, self.patch) = ("", "", "")
        if versionStr == None or versionStr == "": 
            return

        versionTupl = versionStr.split('.')
        if len(versionTupl) > 3: 
            raise Exception("Invalid version. Version is too long")
        try:
            if len(versionTupl) >= 1: 
                self.major = int(versionTupl[0])
            if len(versionTupl) >= 2: 
                self.minor = int(versionTupl[1])
            if len(versionTupl) == 3: 
                self.patch = int(versionTupl[2])
        except ValueError as e:
            raise ValueError("Invalid version. Version may only contain numbers")

    def __hash__(self):
        return self.major * 10000 + self.minor * 100 + self.patch

    def __eq__(self, other):
        return (self.major == self.major and
                self.minor == self.minor and
                self.patch == self.patch)

    def __lt__(self, other):
        if self.major == "" and other.major != "": return True;
        if self.major < other.major: return True;
        if self.minor == "" and other.minor != "": return True;
        if self.minor < other.minor: return True;
        if self.patch == "" and other.patch != "": return True;
        if self.patch < other.patch: return True;

    def __str__(self):
        return f"{str(self.major)}.{str(self.minor)}.{str(self.patch)}"


def __upgradeScript1_0_0(data):
    def fixDiffBlock(data):
        if not isinstance(data, dict): 
            return data
        leftpath = ""
        rightpath = ""
        leftpathkey = None
        rightpathkey = None
        for key in data:
            if key.lower() == "leftpath":
                leftpath = data[key]
                leftpathkey = key
            elif key.lower() == "rightpath":
                rightpath = data[key]
                rightpathkey = key
        if leftpathkey: data.pop(leftpathkey, None)
        if rightpathkey: data.pop(rightpathkey, None)
        if (leftpath == "" and rightpath == ""):
            return data
        data["paths"] = "\"" + leftpath + "\" \"" + rightpath + "\""
        return data
    
    def fixRunBlock(data):
        if isinstance(data, str):
            return {'__line__': -1, 'cmd' : data, 'exitcode': 0}
        if not isinstance(data, dict): 
            return data
        for key in data:
            exitcode = None
            if key.lower() == "exitCode":
                exitcode = data[key]
        if exitcode == None:
            data["exitcode"] = 0
        return data

    def fixLooperData(data):
        for key in data:
            if key.lower() == "steps":
                fixStepsData(data[key])

    def fixStepsData(data):
        if data == None: return
        if not isinstance(data, list): 
            raise Exception("Scriptfile can not be upgraded")

        for step in data:
            for key in step:
                if key.lower() == "steps":
                    fixStepsData(step[key])
                if key.lower() == "run":
                    step[key] = fixRunBlock(step[key])
                if key.lower() == "diff":
                    step[key] = fixDiffBlock(step[key])

    data["version"]="1.1.0"
    fixLooperData(data)
    return (data, Version("1.1.0"))

def __upgradeScript1_1_0(data):
    def fixLooperData(data):
        for key in data:
            if key.lower() == "steps":
                fixStepsData(data[key])

    def fixStepsData(data):
        if data == None: return
        if not isinstance(data, list): 
            raise Exception("Scriptfile can not be upgraded")

        for step in data:
            for key in step:
                if key.lower() == "steps":
                    fixStepsData(step[key])
                if key.lower() == "diff":
                    step[key] = fixDiffBlock(step[key])
                if key.lower() == "run":
                    step[key] = fixRunBlock(step[key])

    def fixRunBlock(data):
        if not isinstance(data, dict): 
            return data
        for key in data:
            if key.lower() == "exitcode":
                if data[key] == 'None':
                    data.pop(key, None)
                break
        return data

    def fixDiffBlock(data):
        if not isinstance(data, dict): 
            return data
        for key in data:
            if key.lower() == "ignorelines":
                data['ignore'] = data[key]
                data.pop(key, None)
                break
        return data

    data["version"]="1.2.0"
    fixLooperData(data)
    return (data, Version("1.2.0"))

__upgradeDic = {Version("1.0.0"): __upgradeScript1_0_0,
                Version("1.1.0"): __upgradeScript1_1_0,}
__workingVersion = Version("1.2.0")

def __upgradeScriptData(data_, printer):
    if data_ == None: return (False, data_)
    data = data_.copy()

    upgraded = False
    currentVersion = Version(data.get("version", str(__workingVersion)))
    if (currentVersion < __workingVersion): 
        printer.verbose(f"Warning: script file needs to be upgraded to version: {str(__workingVersion)}. Use --upgradeScriptFile to save upgraded script file")
        upgraded = True
    while(currentVersion < __workingVersion):
        (data, currentVersion) = __upgradeDic[currentVersion](data)
    return (upgraded, data)

def __removeLineInfo(data):
    if isinstance(data, dict):
        data.pop("__line__", None)
    if isinstance(data, list):
        for item in data:
            if isinstance(item, dict) or isinstance(item, list):
                __removeLineInfo(item)
    if isinstance(data, dict):
        for key in data:
            if isinstance(data[key], dict) or isinstance(data[key], list): 
                __removeLineInfo(data[key])

def processScriptData(file, saveUpgradedScript = False, printer = Printer()):
    with open(file) as f:
        data = yaml.load(f, Loader=SafeLineLoader)
    if data == None: return data
    (upgradedScript, data) = __upgradeScriptData(data, printer)
    if saveUpgradedScript and upgradedScript:
        dumpData = copy.deepcopy(data)
        __removeLineInfo(dumpData)
        with open(file, "w") as f:
            yaml.dump(dumpData, f, sort_keys=False)
    data.pop("version", None)
    return data



import shlex
import subprocess
import os
import pathlib
import sys
import platform
import threading

from mazikeen.ConsolePrinter import Printer
from mazikeen.Utils import replaceVariables, ensure_dir

__pythonPath = None
__pythonPath3 = None
__searchedForPython = False
__searchedForPythonLock = threading.Lock()

class RunBlock:
    def __init__(self, cmd, outputfile = None, inputfile = None, exitcode = None, shell = None):
        self.cmd = cmd
        self.outputfile = outputfile
        self.inputfile = inputfile
        self.exitcode = exitcode
        self.shell = shell
        
    def run(self, workingDir = ".", variables = {}, printer = Printer()):

        printer.verbose("Run:", self.cmd)
        replCmd = replaceVariables(self.cmd, variables)
        printer.verbose("cwd:", os.path.abspath(workingDir))
        printer.verbose("call:", replCmd)
        cmdNArgs = shlex.split(replCmd)

        if self.shell == "powershell": return self.__run_powershell(replCmd, workingDir, variables, printer)
        elif self.shell == "cmd": return self.__run_cmd(replCmd, workingDir, variables, printer)
        elif self.shell == "sh": return self.__run_sh(replCmd, workingDir, variables, printer)
        elif self.shell == "python": return self.__run_pythonX(replCmd, "python", workingDir, variables, printer)
        elif self.shell == "python3": return self.__run_pythonX(replCmd, "python3", workingDir, variables, printer)
        elif self.shell != None: 
            printer.error("unknown shell ", self.shell)
            return false;

        inputfileData = None
        if self.inputfile:
            with open(pathlib.PurePath(workingDir).joinpath(replaceVariables(self.inputfile, variables)), "rb") as fh:
                inputfileData = fh.read()
        shell = (sys.platform == "win32")
        subProcessRes = subprocess.run(cmdNArgs, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, input=inputfileData, cwd = workingDir, shell = shell)
        if self.outputfile:
            outputfileFullPath = str(pathlib.PurePath(workingDir).joinpath(replaceVariables(self.outputfile, variables)))
            ensure_dir(outputfileFullPath)
            with open(outputfileFullPath, "wb") as fh:
                fh.write(subProcessRes.stdout)
        
        res = True
        if (self.exitcode != None):
            res = subProcessRes.returncode == self.exitcode
            if not res:
                printer.error("different exitcode received:", subProcessRes.returncode, "!=", self.exitcode, "for command '"+ str(replCmd) +"'")
        return res

    def __run_powershell(self, replCmd, workingDir = ".", variables = {}, printer = Printer()):
        if self.inputfile:
            printer.error("inputfile not allowed when shell = ", self.shell)
            return false;
        inputData = str.encode(replCmd)
        subProcessRes = subprocess.run(["powershell", "-NonInteractive", "-NoLogo"], stdout=subprocess.PIPE, stderr=subprocess.STDOUT, input=inputData, cwd = workingDir, shell = False)
        if self.outputfile:
            outputfileFullPath = str(pathlib.PurePath(workingDir).joinpath(replaceVariables(self.outputfile, variables)))
            ensure_dir(outputfileFullPath)
            with open(outputfileFullPath, "wb") as fh:
                fh.write(subProcessRes.stdout)

        res = True
        if (self.exitcode != None):
            res = subProcessRes.returncode == self.exitcode
            if not res:
                printer.error("different exitcode received:", subProcessRes.returncode, "!=", self.exitcode, "for command '"+ str(replCmd) +"'")
        return res

    def __run_pythonX(self, replCmd, pythonType = "python3", workingDir = ".", variables = {}, printer = Printer()):
        pythonPath = getPythonPath(pythonType)
        if pythonPath == None:
            printer.error(f"{pythonType} could not be found")
            return false;
        if self.inputfile:
            printer.error("inputfile not allowed when shell = ", self.shell)
            return false;
        inputData = str.encode(replCmd)
        subProcessRes = subprocess.run([pythonPath], stdout=subprocess.PIPE, stderr=subprocess.STDOUT, input=inputData, cwd = workingDir, shell = False)
        if self.outputfile:
            outputfileFullPath = str(pathlib.PurePath(workingDir).joinpath(replaceVariables(self.outputfile, variables)))
            ensure_dir(outputfileFullPath)
            with open(outputfileFullPath, "wb") as fh:
                fh.write(subProcessRes.stdout)

        res = True
        if (self.exitcode != None):
            res = subProcessRes.returncode == self.exitcode
            if not res:
                printer.error("different exitcode received:", subProcessRes.returncode, "!=", self.exitcode, "for command '"+ str(replCmd) +"'")
        return res

    def __run_sh(self, replCmd, workingDir = ".", variables = {}, printer = Printer()):
        if self.inputfile:
            printer.error("inputfile not allowed when shell = ", self.shell)
            return false;
        inputData = str.encode(replCmd)
        subProcessRes = subprocess.run(["sh"], stdout=subprocess.PIPE, stderr=subprocess.STDOUT, input=inputData, cwd = workingDir, shell = False)
        if self.outputfile:
            outputfileFullPath = str(pathlib.PurePath(workingDir).joinpath(replaceVariables(self.outputfile, variables)))
            ensure_dir(outputfileFullPath)
            with open(outputfileFullPath, "wb") as fh:
                fh.write(subProcessRes.stdout)

        res = True
        if (self.exitcode != None):
            res = subProcessRes.returncode == self.exitcode
            if not res:
                printer.error("different exitcode received:", subProcessRes.returncode, "!=", self.exitcode, "for command '"+ str(replCmd) +"'")
        return res

    def __run_cmd(self, replCmd, workingDir = ".", variables = {}, printer = Printer()):
        if self.inputfile:
            printer.error("inputfile not allowed when shell = ", self.shell)
            return false;

        if replCmd.endswith("/n"): 
            inputData = str.encode(replCmd)
        else: 
            inputData = str.encode(replCmd + "\n")
        
        subProcessRes = subprocess.run(["cmd"], stdout=subprocess.PIPE, stderr=subprocess.STDOUT, input=inputData, cwd = workingDir, shell = False)
        if self.outputfile:
            outputfileFullPath = str(pathlib.PurePath(workingDir).joinpath(replaceVariables(self.outputfile, variables)))
            ensure_dir(outputfileFullPath)
            with open(outputfileFullPath, "wb") as fh:
                fh.write(subProcessRes.stdout)

        res = True
        if (self.exitcode != None):
            res = subProcessRes.returncode == self.exitcode
            if not res:
                printer.error("different exitcode received:", subProcessRes.returncode, "!=", self.exitcode, "for command '"+ str(replCmd) +"'")
        return res

def getPythonPath(pythonType):
    with __searchedForPythonLock:
        if __searchedForPython == False:
            if platform.system().lower() == "windows":
                process = subprocess.run(["where", "python"], stdout=subprocess.PIPE, shell = False)
                pythonPaths = process.stdout.decode("utf-8").split("\r\n")
                pythonPaths = list(filter (("").__ne__, pythonPaths))
                for pythonPath in pythonPaths:
                    process = subprocess.run([pythonPath, "--version"], stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell = False)
                    pythonVersion = process.stdout.decode("utf-8")
                    if pythonVersion.startswith("Python 3"):
                        __pythonPath3 = pythonPath
                    elif pythonVersion.startswith("Python 2"):
                        __pythonPath = pythonPath
            if platform.system().lower() == "linux":
                process = subprocess.run(["which", "python"], stdout=subprocess.PIPE, shell = False)
                pythonPaths = process.stdout.decode("utf-8").split("\n")
                process = subprocess.run(["which", "python3"], stdout=subprocess.PIPE, shell = False)
                pythonPaths.extend(process.stdout.decode("utf-8").split("\n"))
                pythonPaths = list(filter (("").__ne__, pythonPaths))
                for pythonPath in pythonPaths:
                    process = subprocess.run([pythonPath, "--version"], stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell = False)
                    pythonVersion = process.stdout.decode("utf-8")
                    if pythonVersion.startswith("Python 3"):
                        __pythonPath3 = pythonPath
                    elif pythonVersion.startswith("Python 2"):
                        __pythonPath = pythonPath
        if pythonType == "python": return __pythonPath
        elif pythonType == "python3": return __pythonPath3
        return None


import re
import os
import shutil
import stat
import pathlib
from enum import Enum

from mazikeen.ConsolePrinter import Printer
from mazikeen.GeneratorException import GeneratorException

__replaceVariablesExp = re.compile(r'(?<!\\)\${.*?}')
def replaceVariables(line, dictReplVar, printer = Printer()):
    global __replaceVariablesExp
    if line == None: return (True, line)
    searchStart = 0
    while (True):
        m = __replaceVariablesExp.search(line[searchStart:])
        if not m: 
            break
        foundVar = m.group()[2:-1]
        replaceVal = dictReplVar.get(foundVar)
        if (replaceVal):
            strReplaceVal = str(replaceVal)
            line = line[:searchStart + m.start()] + strReplaceVal + line[searchStart + m.end():]
            searchStart += m.start() + len(strReplaceVal)
        else:
            raise GeneratorException("Variable " + str(m.group()) + " does not exist")
    line = line.replace(r'\$', "$")
    return line

def ensure_dir(file_path):
    directory = os.path.dirname(file_path)
    if (directory) and (not os.path.exists(directory)):
        os.makedirs(directory)

def rmtree(path, printer = Printer()):
    try:
        def remove_readonly(fn, path, excinfo):
            os.chmod(path, stat.S_IWRITE)
            fn(path)
        if os.path.exists(path):
            shutil.rmtree(path, onerror=remove_readonly, ignore_errors = False)
        return True
    except Exception as e:
        printer.error("rmtree:", e)
        return False

class diffStrategy(Enum):
   IgnoreLeftOrphans = 0
   IgnoreRightOrphans = 1
   IgnoreOrphans = 2
   All = 3

def __listAllFilesWithoutRoot(path):
    allFilesWithRoot = [pathlib.Path(dp) / f for dp, dn, fn in os.walk(os.path.expanduser(path)) for f in fn + dn]
    rootMathLen = len(pathlib.Path(path).parts)
    filesWithoutRoot = [pathlib.Path(*f.parts[rootMathLen:]) for f in allFilesWithRoot]
    return filesWithoutRoot
    
def __getCompareLine(line, fh, compiledignore):
    skipedLines = 0
    while(True):
        if not line: return (line, skipedLines)
        try:
            strLine = line.decode('utf-8')
        except:
            return (line, skipedLines)

        lineMatchesSpan = []
        for ignoreLine in compiledignore:
            itMatch = ignoreLine.finditer(strLine)
            for m in itMatch: lineMatchesSpan.append(m.span())
        lineMatchesSpan.sort()

        def concatenatelineMatches(lineMatchesSpan):
            idx = 0
            while idx < len(lineMatchesSpan) - 1:
                cmpIdx = idx + 1
                while cmpIdx < len(lineMatchesSpan):
                    if lineMatchesSpan[idx][1] > lineMatchesSpan[cmpIdx][0]:
                        lineMatchesSpan[idx] = (lineMatchesSpan[idx][0], max(lineMatchesSpan[idx][1], lineMatchesSpan[cmpIdx][1]))
                        cmpIdx = idx + 1
                        del lineMatchesSpan[cmpIdx]
                    else: cmpIdx += 1
                idx += 1
        concatenatelineMatches(lineMatchesSpan)

        def deleteLineMatchesFromString(strLine, lineMatchesSpan):
            listLine = list(strLine)
            for matchSpan in reversed(lineMatchesSpan):
                del listLine[matchSpan[0]:matchSpan[1]]
            return "".join(listLine)
        newLine = deleteLineMatchesFromString(strLine, lineMatchesSpan)
        if not newLine in ['', '\n', '\r', '\r\n']:
            return (str.encode(newLine), skipedLines)
        line = fh.readline()
        skipedLines += 1
    return (line, skipedLines)

def normalizeEOL(line):
    # Todo: can be optimized
    if len(line) >=2 and line[-2:] == bytes('\r\n', 'utf-8'): 
        line = line[:-2] + bytes('\n', 'utf-8')
    elif len(line) >=1 and line[-1] == bytes('\r', 'utf-8'): 
        line = line[:-1] + bytes('\n', 'utf-8')
    return line

def diffFiles(fileL, fileR, compiledignore = [], binaryCompare = False):
    def areLinesIdentical(lineL, lineR, binaryCompare):
        def normalizeEOL(line):
            if len(line) >=2 and line[-2:] == bytes('\r\n', 'utf-8'): 
                line = line[:-2] + bytes('\n', 'utf-8')
            elif len(line) >=1 and line[-1] == bytes('\r', 'utf-8'): 
                line = line[:-1] + bytes('\n', 'utf-8')
            return line

        if lineL == lineR: return True #This is the most likely scenario so it has priority
        if binaryCompare == False:
            return normalizeEOL(lineL) == normalizeEOL(lineR)
        return False

    with open(fileL, "rb") as fhL:
        with open(fileR, "rb") as fhR:
            (lineNrL, lineNrR) = (0, 0)
            while True:
                lineL = fhL.readline()
                lineNrL += 1
                lineR = fhR.readline()
                lineNrR += 1
                if (not lineL and not lineR): break #EOF reached
                if areLinesIdentical(lineL, lineR, binaryCompare) == False:
                    (lineL, skipedLines) = __getCompareLine(lineL, fhL, compiledignore)
                    lineNrL += skipedLines
                    (lineR, skipedLines) = __getCompareLine(lineR, fhR, compiledignore)
                    lineNrR += skipedLines
                    if areLinesIdentical(lineL, lineR, binaryCompare) == False:
                        return {"lineNrL": lineNrL, 
                                "lineNrR": lineNrR, 
                                "lineL": lineL,
                                "lineR": lineR}
    return None

def diff(pathL, pathR, binaryCompare = False, diffStrategy = diffStrategy.All, ignore = [], printer = Printer(verbose = True)):
    rootM = pathlib.Path(pathL)
    rootS = pathlib.Path(pathR)
    compiledignore = list(map(lambda x: re.compile(x), ignore))
    if (rootM.is_file() and rootS.is_file()):
        diffRes = diffFiles(rootM, rootS, compiledignore = compiledignore, binaryCompare = binaryCompare)
        if (diffRes):
            if printer.isVerbose():
                printer.error(f"diff failed: '{rootM}' != '{rootS}'\nwhere: {diffRes['lineNrL']}: {diffRes['lineL']} != {diffRes['lineNrR']}: {diffRes['lineR']}")
            else:
                printer.error(f"diff failed: '{rootM}' != '{rootS}'")
            return False
        return True

    for file in [rootM, rootS]:
        if not file.exists():
            printer.error(f"diff failed: '{file}' doesn't exist")
            return False

    filesM = set(__listAllFilesWithoutRoot(pathL))
    filesS = set(__listAllFilesWithoutRoot(pathR))
    
    swapFiles = False
    if diffStrategy == diffStrategy.IgnoreLeftOrphans:
        swapFiles = True
        filesM, filesS = filesS, filesM
        rootM, rootS = rootS, rootM
    
    if diffStrategy != diffStrategy.IgnoreOrphans:
        for file in (filesM - filesS):
            printer.error(f"diff failed: '{str(rootM/file)}' not in '{rootS / file.parent}'")
            return False
    
    if diffStrategy == diffStrategy.All:
        for file in (filesS - filesM):
            printer.error(f"diff failed: '{str(rootS/file)}' not in '{rootM / file.parent}'")
            return False
    
    comPaths =[(rootM/path, rootS/path) for path in (filesM & filesS)]
    
    for pathM, pathS in comPaths:
        if (pathM.is_file() != pathS.is_file()):
            (pathL_, pathR_) = (pathM, pathS) if not swapFiles else (pathS, pathM)
            printer.error(f"diff failed: '{pathL_}' != '{pathR_}'")
            return False

    comFiles = [(pathM, pathS) for pathM, pathS in comPaths if pathM.is_file()]
    for fileM, fileS in comFiles:
        diffRes = diffFiles(fileM, fileS, binaryCompare = binaryCompare, compiledignore = compiledignore)
        if (diffRes):
            (fileL, fileR) = (fileM, fileS) if not swapFiles else (fileS, fileM)
            if printer.isVerbose():
                printer.error(f"diff failed: '{fileL}' != '{fileR}'\nwhere: {diffRes['lineNrL']}: {diffRes['lineL']} != {diffRes['lineNrR']}: {diffRes['lineR']}")
            else:
                printer.error(f"diff failed: '{fileL}' != '{fileR}'")
            return False
    return True




from mazikeen.MakedirsBlock import MakedirsBlock
from mazikeen.GeneratorException import GeneratorException

def generateMakedirs(data):
    if not isinstance(data, str):
        raise GeneratorException("'makedirs' block not recognized")
    return MakedirsBlock(data)


import unittest
import subprocess
import os
import re
import platform
from mazikeen.Utils import diff, rmtree, diffStrategy
from xmldiff import main
from distutils.dir_util import copy_tree

class Blackbox(unittest.TestCase):
    def test_simple(self):
        testDir = "TestFiles/Blackbox_test/simple/"
        outDir = testDir + "TestOutput"
        rmtree(outDir)
        os.makedirs(outDir)
        with open(outDir + "/mazikenout.txt", "w") as of:
            subprocess.run(["mazikeen"], stdout=of, stderr=of, cwd = testDir)
        self.assertTrue(diff(testDir + "TestOutput/mazikenout.txt", testDir + "TestExpected/mazikenout.txt", ignore = ["process time: .*"]))
        
    def test_cwd(self):
        testDir = "TestFiles/Blackbox_test/cwd/"
        outDir = testDir + "TestOutput"
        rmtree(outDir)
        os.makedirs(outDir)
        with open(outDir + "/mazikenout.txt", "w") as of:
            subprocess.run(["mazikeen", "-v"], stdout=of, stderr=of, cwd = testDir)
        self.assertTrue(diff(testDir + "TestOutput/mazikenout.txt", testDir + "TestExpected/mazikenout.txt", ignore = ["process time: .*", "cwd: .*"]))
        
    def test_cmdArg_ScriptName(self):
        testDir = "TestFiles/Blackbox_test/cmdArg_ScriptName/"
        outDir = testDir + "TestOutput"
        rmtree(outDir)
        os.makedirs(outDir)
        with open(outDir + "/mazikenout.txt", "w") as of:
            subprocess.run(["mazikeen", "--scriptName", "dummyScriptName.yaml"], stdout=of, stderr=of, cwd = testDir)
        self.assertTrue(diff(testDir + "TestOutput/mazikenout.txt", testDir + "TestExpected/mazikenout.txt", ignore = ["process time: .*"]))

    def test_inputFileNoutputFile(self):
        testDir = "TestFiles/Blackbox_test/inputFileNoutputFile/"
        outDir = testDir + "TestOutput"
        rmtree(outDir)
        os.makedirs(outDir)
        with open(outDir + "/mazikenout.txt", "w") as of:
            subprocess.run(["mazikeen"], stdout=of, stderr=of, cwd = testDir)
        self.assertTrue(diff(testDir + "TestOutput/mazikenout.txt", testDir + "TestExpected/mazikenout.txt", ignore = ["process time: .*"]))
        
    def test_blockinBlock(self):
        testDir = "TestFiles/Blackbox_test/blockinBlock/"
        outDir = testDir + "TestOutput"
        rmtree(outDir)
        os.makedirs(outDir)
        with open(outDir + "/mazikenout.txt", "w") as of:
            subprocess.run(["mazikeen"], stdout=of, stderr=of, cwd = testDir)
        self.assertTrue(diff(testDir + "TestOutput/mazikenout.txt", testDir + "TestExpected/mazikenout.txt", ignore = ["process time: .*"]))
        
    def test_testsuitsNtestcases_simple(self):
        testDir = "TestFiles/Blackbox_test/testsuitsNtestcases/"
        outDir = testDir + "TestOutput/simple/"
        rmtree(outDir)
        os.makedirs(outDir)
        with open(outDir + "/mazikenout.txt", "w") as of:
            subprocess.run(["mazikeen"], stdout=of, stderr=of, cwd = testDir)
        self.assertTrue(diff(testDir + "TestOutput/simple/mazikenout.txt", testDir + "TestExpected/simple/mazikenout.txt", ignore = ["process time: .*"]))
        
    def test_testsuitsNtestcases_report(self):
        testDir = "TestFiles/Blackbox_test/testsuitsNtestcases/"
        outDir = testDir + "TestOutput/report"
        rmtree(outDir)
        os.makedirs(outDir)
        with open(outDir + "/mazikenout.txt", "w") as of:
            subprocess.run(["mazikeen", "-r", "TestOutput/report/report.xml"], stdout=of, stderr=of, cwd = testDir)
        
        with open(outDir + "/report.xml", "r") as ifile:
            with open(outDir + "/report_diff.xml", "w") as ofile:
                for line in ifile:
                    line = re.sub(r"time=\".+?\"", "time=\"\"", line)
                    line = re.sub(r"\\", "/", line)
                    ofile.write(line)
        self.assertTrue(diff(testDir + "TestOutput/report/report_diff.xml", testDir + "TestExpected/report/report_diff.xml", ignore = ['time=""', " "]))
        
    def test_testsuitsNtestcases_wait_parallel(self):
        testDir = "TestFiles/Blackbox_test/testsuitsNtestcases_wait/"
        outDir = testDir + "TestOutput/parallel/"
        rmtree(outDir)
        os.makedirs(outDir)
        with open(outDir + "/mazikenout.txt", "w") as of:
            subprocess.run(["mazikeen", "-r", "TestOutput/parallel/report.xml", "-j", "2"], stdout=of, stderr=of, cwd = testDir)
        
        with open(outDir + "/report.xml", "r") as ifile:
            with open(outDir + "/report_diff.xml", "w") as ofile:
                for line in ifile:
                    line = re.sub(r"time=\".+?\"", "time=\"\"", line)
                    line = re.sub(r"\\", "/", line)
                    ofile.write(line)
        self.assertEqual(main.diff_files(outDir + "/report_diff.xml", outDir + "/report_diff.xml"), [])
        self.assertTrue(diff(testDir + "TestOutput/parallel/mazikenout.txt", testDir + "TestExpected/parallel/mazikenout.txt", ignore = ["process time: .*"]))
        
    def test_testsuitsNtestcases_waitNfail_parallel(self):
        testDir = "TestFiles/Blackbox_test/testsuitsNtestcases_waitNfail/"
        outDir = testDir + "TestOutput/parallel/"
        rmtree(outDir)
        os.makedirs(outDir)
        with open(outDir + "/mazikenout.txt", "w") as of:
            subprocess.run(["mazikeen", "-r", "TestOutput/parallel/report.xml", "-j", "2"], stdout=of, stderr=of, cwd = testDir)
        
        with open(outDir + "/report.xml", "r") as ifile:
            with open(outDir + "/report_diff.xml", "w") as ofile:
                for line in ifile:
                    line = re.sub(r"time=\".+?\"", "time=\"\"", line)
                    line = re.sub(r"\\", "/", line)
                    ofile.write(line)
        self.assertEqual(main.diff_files(outDir + "/report_diff.xml", outDir + "/report_diff.xml"), [])
        self.assertTrue(diff(testDir + "TestOutput/parallel/mazikenout.txt", testDir + "TestExpected/parallel/mazikenout.txt", ignore = ["process time: .*"]))

    def test_testsuitsNtestcases_waitNfail_serial(self):
        testDir = "TestFiles/Blackbox_test/testsuitsNtestcases_waitNfail/"
        outDir = testDir + "TestOutput/serial/"
        rmtree(outDir)
        os.makedirs(outDir)
        with open(outDir + "/mazikenout.txt", "w") as of:
            subprocess.run(["mazikeen", "-r", "TestOutput/serial/report.xml", "-j", "2"], stdout=of, stderr=of, cwd = testDir)
        
        with open(outDir + "/report.xml", "r") as ifile:
            with open(outDir + "/report_diff.xml", "w") as ofile:
                for line in ifile:
                    line = re.sub(r"time=\".+?\"", "time=\"\"", line)
                    line = re.sub(r"\\", "/", line)
                    ofile.write(line)
        self.assertEqual(main.diff_files(outDir + "/report_diff.xml", outDir + "/report_diff.xml"), [])
        self.assertTrue(diff(testDir + "TestOutput/serial/mazikenout.txt", testDir + "TestExpected/serial/mazikenout.txt", ignore = ["process time: .*"]))

    def test_testsuitsNtestcases_waitNfail_failFast(self):
        testDir = "TestFiles/Blackbox_test/testsuitsNtestcases_waitNfail/"
        outDir = testDir + "TestOutput/failFast/"
        rmtree(outDir)
        os.makedirs(outDir)
        with open(outDir + "/mazikenout.txt", "w") as of:
            subprocess.run(["mazikeen", "-r", "TestOutput/failFast/report.xml", "-j", "2", "--failfast"], stdout=of, stderr=of, cwd = testDir)
        
        with open(outDir + "/report.xml", "r") as ifile:
            with open(outDir + "/report_diff.xml", "w") as ofile:
                for line in ifile:
                    line = re.sub(r"time=\".+?\"", "time=\"\"", line)
                    line = re.sub(r"\\", "/", line)
                    ofile.write(line)
        self.assertEqual(main.diff_files(outDir + "/report_diff.xml", outDir + "/report_diff.xml"), [])
        self.assertTrue(diff(testDir + "TestOutput/failFast/mazikenout.txt", testDir + "TestExpected/failFast/mazikenout.txt", ignore = ["process time: .*"]))

    def test_upgradeScriptData1_0_0(self):
        testDir = "TestFiles/Blackbox_test/upgradeScriptData1.0.0/"
        outDir = testDir + "TestOutput/"
        rmtree(outDir)
        os.makedirs(outDir)
        copy_tree(testDir+"TestInput", outDir)
        with open(outDir + "/mazikenout.txt", "w") as of:
            subprocess.run(["mazikeen", "--upgradeScriptFile"], stdout=of, stderr=of, cwd = outDir)

        self.assertTrue(diff(testDir + "TestOutput/mazikenout.txt", testDir + "TestExpected/mazikenout.txt", ignore = ["process time: .*"]))
        self.assertTrue(diff(testDir + "TestOutput/script.yaml", testDir + "TestExpected/script.yaml"))

    def test_upgradeScriptData1_1_0(self):
        testDir = "TestFiles/Blackbox_test/upgradeScriptData1.1.0/"
        outDir = testDir + "TestOutput/"
        rmtree(outDir)
        os.makedirs(outDir)
        copy_tree(testDir+"TestInput", outDir)
        with open(outDir + "/mazikenout.txt", "w") as of:
            subprocess.run(["mazikeen", "--upgradeScriptFile"], stdout=of, stderr=of, cwd = outDir)

        self.assertTrue(diff(testDir + "TestOutput/mazikenout.txt", testDir + "TestExpected/mazikenout.txt", ignore = ["process time: .*"]))
        self.assertTrue(diff(testDir + "TestOutput/script.yaml", testDir + "TestExpected/script.yaml"))

    def test_emptyTest(self):
        testDir = "TestFiles/Blackbox_test/emptyTest/"
        outDir = testDir + "TestOutput/"
        rmtree(outDir)
        os.makedirs(outDir)
        with open(outDir + "/mazikenout.txt", "w") as of:
            subprocess.run(["mazikeen"], stdout=of, stderr=of, cwd = testDir)
        self.assertTrue(diff(testDir + "TestOutput/mazikenout.txt", testDir + "TestExpected/mazikenout.txt", ignore = ["process time: .*"]))

    def test_shellWindows(self):
        if (platform.system() == "Windows"):
            testDir = "TestFiles/Blackbox_test/shellWin/"
            outDir = testDir + "TestOutput"
            rmtree(outDir)
            os.makedirs(outDir)
            with open(outDir + "/mazikenout.txt", "w") as of:
                subprocess.run(["mazikeen"], stdout=of, stderr=of, cwd = testDir)
            self.assertTrue(diff(testDir + "TestOutput/mazikenout.txt", testDir + "TestExpected/mazikenout.txt", ignore = ["process time: .*"]))

    def test_shellLinux(self):
        if (platform.system() == "Linux"):
            testDir = "TestFiles/Blackbox_test/shellLinux/"
            outDir = testDir + "TestOutput"
            rmtree(outDir)
            os.makedirs(outDir)
            with open(outDir + "/mazikenout.txt", "w") as of:
                subprocess.run(["mazikeen"], stdout=of, stderr=of, cwd = testDir)
            self.assertTrue(diff(testDir + "TestOutput/mazikenout.txt", testDir + "TestExpected/mazikenout.txt", ignore = ["process time: .*"]))

    def test_shellPython(self):
        testDir = "TestFiles/Blackbox_test/shellPython/"
        outDir = testDir + "TestOutput"
        rmtree(outDir)
        os.makedirs(outDir)
        with open(outDir + "/mazikenout.txt", "w") as of:
            subprocess.run(["mazikeen"], stdout=of, stderr=of, cwd = testDir)
        self.assertTrue(diff(testDir + "TestOutput/mazikenout.txt", testDir + "TestExpected/mazikenout.txt", ignore = ["process time: .*"]))

    
if __name__ == '__main__':
    unittest.main()



import unittest
import os
import io
import sys
from mazikeen.RunBlock import RunBlock
from mazikeen.ConsolePrinter import Printer
from mazikeen.Utils import rmtree

class RunBlockTest(unittest.TestCase):
    def test_basic(self):
        cmdExe = RunBlock(r"python -c \"print('Hello World')\"")
        res = cmdExe.run()
        self.assertEqual(res, True)
        
    def test_basicVerbose(self):
        capturedOutput = io.StringIO()
        sys.stdout = capturedOutput
        printer = Printer(verbose = True)
        cmdString = r"python -c \"print('Hello World')\""
        cmdExe = RunBlock(cmdString)
        res = cmdExe.run(printer = printer)
        sys.stdout = sys.__stdout__
        expectedResult = 'Run: python -c \\"print(\'Hello World\')\\"' + "\n" + "cwd: " + os.getcwd() + "\n" + 'call: '+ cmdString + "\n"
        
        self.assertEqual(res, True)
        self.assertEqual(expectedResult, capturedOutput.getvalue())
    
    def test_unexpectedExitCodeVerbose(self):
        printer = Printer(verbose = True)
        capturedOutput = io.StringIO()
        sys.stdout = capturedOutput
        cmdString = '"' + sys.executable + '"'  + ' -c "exit(1)"'
        cmdExe = RunBlock(cmdString, exitcode = 0)
        res = cmdExe.run(printer = printer)
        sys.stdout = sys.__stdout__
        expectedResult = 'Run: ' + cmdString + "\n" + "cwd: " + os.getcwd() + "\n" + 'call: '+ cmdString + "\n" + 'Error: different exitcode received: 1 != 0 for command \'' + cmdString + "\'\n"
        
        self.assertEqual(res, False)
        resStr = capturedOutput.getvalue()
        self.assertEqual(expectedResult, capturedOutput.getvalue())
        
    def test_exitCodeNone(self):
        printer = Printer(verbose = True)
        capturedOutput = io.StringIO()
        sys.stdout = capturedOutput
        cmdString = '"' + sys.executable + '"'  + ' -c "exit(1)"'
        cmdExe = RunBlock(cmdString)
        res = cmdExe.run(printer = printer)
        sys.stdout = sys.__stdout__
        expectedResult ='Run: ' + cmdString + "\n" + "cwd: " + os.getcwd() + "\n" + 'call: '+ cmdString + "\n"
        
        self.assertEqual(res, True)
        resStr = capturedOutput.getvalue()
        self.assertEqual(expectedResult, capturedOutput.getvalue())
        
    def test_inputfileNoutputfile(self):
        rmtree("Output")
        printer = Printer(verbose = True)
        
        capturedOutput = io.StringIO()
        sys.stdout = capturedOutput
        
        cmdExe = RunBlock('"' + sys.executable + '"' + " RunBlock_test/stdinTostdout.py", inputfile="RunBlock_test/stdinTostdout.py", outputfile="RunBlock_test/Output/inputfileNoutputfile.txt")
        res = cmdExe.run(workingDir = "TestFiles",printer = printer)
        sys.stdout = sys.__stdout__
        
        self.assertEqual(res, True)
        with open("TestFiles/RunBlock_test/stdinTostdout.py", "r") as fh1:
            with open("TestFiles/RunBlock_test/Output/inputfileNoutputfile.txt", "r") as fh2:
                self.assertEqual(fh1.read(), fh2.read())
        rmtree("TestFiles/RunBlock_test/Output")
    def test_variables(self):
        rmtree("Output")
        printer = Printer(verbose = True)
        
        capturedOutput = io.StringIO()
        sys.stdout = capturedOutput
        
        cmdExe = RunBlock('"' + sys.executable + '"' + " RunBlock_test/stdinTostdout.py", inputfile="RunBlock_test/${var}.py", outputfile="RunBlock_test/Output/${var}Test.txt")
        res = cmdExe.run(workingDir = "TestFiles",printer = printer, variables = {'var':'stdinTostdout'})
        sys.stdout = sys.__stdout__
        
        self.assertEqual(res, True)
        with open("TestFiles/RunBlock_test/stdinTostdout.py", "r") as fh1:
            with open("TestFiles/RunBlock_test/Output/stdinTostdoutTest.txt", "r") as fh2:
                self.assertEqual(fh1.read(), fh2.read())
        rmtree("TestFiles/RunBlock_test/Output")

if __name__ == '__main__':
    unittest.main()




import setuptools
import mazikeen.version

with open("README.rst", "r", encoding="utf-8") as fh:
    long_description = fh.read()
    
with open("LICENSE", "r", encoding="utf-8") as fh:
    LICENSE = fh.read()

setuptools.setup(
    name="mazikeen",
    version=mazikeen.version.__version__,
    author="Neaga Septimiu",
    author_email="neagas@gmail.com",
    description="Test enviroment for CLI application",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/hanniballar/mazikeen",
    project_urls={
        "Bug Tracker": "https://github.com/hanniballar/mazikeen/issues",
    },
    license_file = LICENSE,
    license="MIT",
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Testing",
        "Topic :: Utilities"
    ],
    packages=["mazikeen"],
    install_requires=["junit_xml>=1.8", "pyyaml>=5.4.1"],
    extras_require={
        'testing': [
            "xmldiff==2.4"
        ]
    },
    entry_points={"console_scripts": ["mazikeen=mazikeen.__main__:main"]},
)


import unittest
import yaml
import re
from mazikeen.Loopers import Serial, Parallel
from mazikeen.GeneratorLooper import generateSerialBlock, generateParallelBlock
from mazikeen.GeneratorException import GeneratorException
from mazikeen.ScriptDataProcessor import SafeLineLoader

class GeneratorSerialTest(unittest.TestCase):
    def test_basic(self):
        with open('TestFiles/GeneratorLooper_test/test_basic/script.yaml') as f:
            data = yaml.load(f, Loader=SafeLineLoader)
            block = generateSerialBlock(data)
            self.assertEqual(block.entries, [{'color': 'red', 'shape': 'circle'}, {'color': 'yellow', 'shape': 'square'}, {'shape': 'triangle'}])
            self.assertEqual(block.failfast, False)
            self.assertEqual(block.steps[0].steps, [])
            self.assertEqual(block.steps[0].entries, [{'color': 'red', 'shape': 'circle'}, {'color': 'yellow', 'shape': 'circle'}, {'color': 'red', 'shape': 'square'}, {'color': 'yellow', 'shape': 'square'}, {'color': 'red', 'shape': 'triangle'}, {'color': 'yellow', 'shape': 'triangle'}])
            self.assertEqual(block.steps[0].steps, [])

    def test_multipleEntries(self):
        with open('TestFiles/GeneratorLooper_test/test_multipleEntries/script.yaml') as f:
            data = yaml.load(f, Loader=SafeLineLoader)
            block = generateSerialBlock(data)
            self.assertEqual(block.entries, [{'color': 'red', 'shape': 'circle'}, {'color': 'yellow', 'shape': 'square'}, {'shape': 'triangle'}, {'fruit': 'apple', 'vegetable': 'onion'}, {'fruit': 'apple', 'vegetable': 'carot'}, {'fruit': 'banana', 'vegetable': 'onion'}, {'fruit': 'banana', 'vegetable': 'carot'}])
            self.assertEqual(block.steps, [])
    def test_invalid_field(self):
        with open('TestFiles/GeneratorLooper_test/test_invalid_field/script.yaml') as f:
            data = yaml.load(f, Loader=SafeLineLoader)
            with self.assertRaisesRegex(GeneratorException, "Only one of the following keys are allowed: .*") as e:
                block = generateSerialBlock(data)
    def test_invalid_entries_field(self):
        with open('TestFiles/GeneratorLooper_test/test_invalid_entries_field/script.yaml') as f:
            data = yaml.load(f, Loader=SafeLineLoader)
            with self.assertRaisesRegex(GeneratorException, "Only one of the following keys are allowed: .*") as e:
                block = generateSerialBlock(data)
    def test_parallel(self):
        with open('TestFiles/GeneratorLooper_test/test_parallel/script.yaml') as f:
            data = yaml.load(f, Loader=SafeLineLoader)
            block = generateSerialBlock(data)
            self.assertTrue(isinstance(block.steps[0], Parallel))
            self.assertEqual(block.failfast, False)
            self.assertEqual(block.steps[0].entries, [{'color': 'red', 'shape': 'circle'}, {'color': 'yellow', 'shape': 'circle'}, {'color': 'red', 'shape': 'square'}, {'color': 'yellow', 'shape': 'square'}, {'color': 'red', 'shape': 'triangle'}, {'color': 'yellow', 'shape': 'triangle'}])
            self.assertEqual(block.steps[0].max_workers, 5)
            self.assertTrue(isinstance(block.steps[0].steps[0], Serial))
    def test_parallel_invalid_max_workers(self):
        with open('TestFiles/GeneratorLooper_test/test_parallel_invalid_max_workers/script.yaml') as f:
            data = yaml.load(f, Loader=SafeLineLoader)
            with self.assertRaisesRegex(GeneratorException, "field 'max_workers' expects an integer at line .*") as e:
                block = generateSerialBlock(data)

if __name__ == '__main__':
    unittest.main()



import unittest
import pathlib
from mazikeen.MakedirsBlock import MakedirsBlock
from mazikeen.ConsolePrinter import Printer
from mazikeen.Utils import rmtree


class MakedirsBlockTest(unittest.TestCase):
    def test_basic(self):
        testPath = pathlib.Path('Output/MkdirBlockTest/TestDir1')
        block = MakedirsBlock(str(testPath))
        res = block.run()
        self.assertEqual(res, True)
        self.assertTrue(testPath.is_dir())
        rmtree('Output')

    def test_workingDir(self):
        testPath = pathlib.Path('MkdirBlockTest/TestDir1')
        block = MakedirsBlock(str(testPath))
        testPath = pathlib.Path("Output").joinpath(testPath)
        res = block.run(workingDir = "Output")
        self.assertEqual(res, True)
        self.assertTrue(testPath.is_dir())
        rmtree('Output')
    
    def test_basic_withTailoringSlash(self):
        testPath = pathlib.Path('Output/MkdirBlockTest/TestDir1/')
        block = MakedirsBlock(str(testPath))
        res = block.run()
        self.assertEqual(res, True)
        self.assertTrue(testPath.is_dir())
        rmtree('Output')
if __name__ == '__main__':
    unittest.main()

